package de.desy.language.snl.ui.compiler;

import java.io.File;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.Assert;

import de.desy.language.snl.ui.compiler.CompileException.CompilerType;
import de.desy.language.snl.ui.preferences.CompilerOptionsService;

public class SNLCompiler {

	private static final Logger logger = Logger.getLogger(SNLCompiler.class
			.getName());

	public static void compile(IFile sourceFile) throws CompileException {
		File compiledSTFile = compileST(sourceFile);

		if (compiledSTFile != null && compiledSTFile.exists()) {
			compileC(sourceFile, compiledSTFile);
		}

		// XXX for tests only
		File srcFile = sourceFile.getLocation().toFile();
		String fileName = srcFile.getName();
		String newFileName = fileName.replace(".st", ".c");
		String parentPath = srcFile.getParent();
		String newPath = parentPath.replace("source", "generated-c");

		String newCFile = newPath + File.separator + newFileName;

		File file = new File(newCFile);
		compileC(sourceFile, file);

	}

	private static File compileST(IFile sourceRessource)
			throws CompileException {

		try {

			logger.info("Start SN-compilation");
			File snlCompilerPath = CompilerOptionsService.defaultInstance()
					.getSNLCompilerPath();

			if (snlCompilerPath == null || !snlCompilerPath.exists()) {
				IMarker errorMarker = sourceRessource
						.createMarker(IMarker.PROBLEM);
				errorMarker.setAttribute(IMarker.SEVERITY,
						IMarker.SEVERITY_ERROR);
				errorMarker
						.setAttribute(
								IMarker.MESSAGE,
								"No SN-Compiler preferences present. Please update your preferences in category SNL / Compiler!");

				logger.severe("No SN-Compiler found");
				return null;
			}

			File source = sourceRessource.getLocation().toFile();
			String fullQualifiedSourceFileName = source.getAbsolutePath();

			File parentPathOfSourcesParent = source.getParentFile()
					.getParentFile();
			String sourceFileName = source.getName();

			String fullQualifiedTargetSourceName = parentPathOfSourcesParent
					.getAbsolutePath()
					+ File.separator
					+ "generated-c"
					+ File.separator
					+ sourceFileName + ".c";

			File targetSourceFile = new File(fullQualifiedTargetSourceName);
			if (targetSourceFile.exists()) {
				System.out
						.println("File deleted? " + targetSourceFile.delete());
			} else {
				System.out.println("No old source "
						+ targetSourceFile.getAbsolutePath());
			}

			String sncCommand = snlCompilerPath.getAbsolutePath()
					+ File.separator + "snc";
			ProcessBuilder processBuilder = new ProcessBuilder(sncCommand,
					"-o ", fullQualifiedTargetSourceName,
					fullQualifiedSourceFileName);
			processBuilder.redirectErrorStream(true);

			Process sncProcess;
			sncProcess = processBuilder.start();
			InputStream stdOut = sncProcess.getInputStream();
			InputStream stdErr = sncProcess.getErrorStream();
			int result = sncProcess.waitFor();
			int c;
			StringBuffer stdOutBuffer = new StringBuffer();
			while ((c = stdOut.read()) != -1) {
				stdOutBuffer.append((char) c);
			}
			String stdOutResult = stdOutBuffer.toString();
			System.out.println("STDOUT:\n" + stdOutResult);
			StringBuffer stdErrBuffer = new StringBuffer();
			while ((c = stdErr.read()) != -1) {
				stdErrBuffer.append((char) c);
			}
			String stdErrResult = stdErrBuffer.toString();
			System.out.println("STDERR:" + stdErrResult);

			if (result != 0) {
				// create marker for SN-Compiler errors
				Pattern pattern = Pattern
						.compile("(syntax error: line no. )([\\d]*)([^\\n]*)(\\n)([\\S\\s]*)");
				Matcher resultMatcher = pattern.matcher(stdOutResult);
				resultMatcher.find();

				IMarker errorMarker = sourceRessource
						.createMarker(IMarker.PROBLEM);
				errorMarker.setAttribute(IMarker.SEVERITY,
						IMarker.SEVERITY_ERROR);
				errorMarker.setAttribute(IMarker.LINE_NUMBER, Integer
						.parseInt(resultMatcher.group(2).trim()));
				errorMarker.setAttribute(IMarker.MESSAGE, resultMatcher
						.group(5).trim());
			} else {
				return targetSourceFile;
			}

		} catch (Exception e) {
			throw new CompileException(CompilerType.SN_Compiler, e);
		}

		return null;
	}

	private static void compileC(IFile sourceRessource, File sourceCFile)
			throws CompileException {
		Assert.isNotNull(sourceRessource);
		Assert.isTrue(sourceRessource.exists());

		try {

			File compilerPath = CompilerOptionsService.defaultInstance()
					.getCCompilerPath();

			if (compilerPath == null || !compilerPath.exists()) {
				IMarker errorMarker = sourceRessource
						.createMarker(IMarker.PROBLEM);
				errorMarker.setAttribute(IMarker.SEVERITY,
						IMarker.SEVERITY_ERROR);
				errorMarker
						.setAttribute(
								IMarker.MESSAGE,
								"No C-Compiler preferences present. Please update your preferences in category SNL / Compiler!");
				return;
			}

			// File sourceFile = sourceRessource.getLocation().toFile();
			File sourceFile = sourceCFile;

			// File compilerPath = new File("/usr/bin/gcc");
			if (compilerPath != null && compilerPath.exists()) {
				logger.info("C-Compiler found");

				List<String> arguments = new ArrayList<String>();
				String compiler = compilerPath.getAbsolutePath();
				String compilerArgument = "-o";

				// create compiled-filename
				String outputFilename = sourceFile.getName().substring(0,
						sourceFile.getName().length() - 2)
						+ ".o";

				// move compiled classes to bin directory
				String oldOutputFile = sourceFile.getParent();
				String newOutputFile = oldOutputFile.replace("generated-c",
						"bin");
				String outputFile = newOutputFile + File.separator
						+ outputFilename;
				String inputFile = sourceFile.getPath();

				arguments.add(compiler);
				arguments.add(compilerArgument);
				arguments.add(outputFile);
				arguments.add(inputFile);

				ProcessBuilder processBuilder = new ProcessBuilder(arguments);
				processBuilder.redirectErrorStream(true);
				Process cprocess = processBuilder.start();

				InputStream stdOut = cprocess.getInputStream();
				InputStream stdErr = cprocess.getErrorStream();

				cprocess.waitFor();
				logger.info("Class compiled");

				// print compile results
				int c;
				StringBuffer stdOutBuffer = new StringBuffer();
				while ((c = stdOut.read()) != -1) {
					stdOutBuffer.append((char) c);
				}
				String stdOutResult = stdOutBuffer.toString();
				System.out.println("STDOUT:\n" + stdOutResult);

				// print compile errors
				StringBuffer stdErrBuffer = new StringBuffer();
				while ((c = stdErr.read()) != -1) {
					stdErrBuffer.append((char) c);
				}
				String stdErrResult = stdErrBuffer.toString();
				System.out.println("STDERR:" + stdErrResult);
			} else {
				logger.log(Level.SEVERE, "C-Compiler could not be found");
			}

		} catch (Exception e) {
			throw new CompileException(CompilerType.C_Compiler, e);
		}

	}
}
